# cython: language_level=3

cdef enum SIGNALS:
    SHOUT, WHISPER, JOIN, LEAVE, PEERS, PEERS_BY_GROUP, OWN_GROUPS, PEER_GROUPS, PEER_ADDRESS, PEER_HEADER_VALUE

cdef const char * TERMINATE
cdef const char * INCOMING